
type atom = string
type sym = string
type var = Var.var
type atom_sort = string
type data_sort = string

type sort = AtomSt of atom_sort
  | DataSt of data_sort
  | UnitSt
  | PairSt of sort * sort
  | AbsSt of string * sort
  | ListSt of sort

type arity = sort * data_sort

type sym_kind = AtomSK 
  | DataSK 
  | AtomK of atom_sort 
  | SymK of arity 
  | PropK of sort

type sg = (string * sym_kind) list


module AtomMap = Map.Make(struct type t = atom let compare = compare end)
type perm = (atom*atom) list
let id = []
let transpose a b = [(a,b)]
let compose p q = p @ q
let mk_perm l = l


type term = Atom of atom
  | Unit   
  | Pair of term * term
  | App of sym * term
  | Abs of atom * term
  | Susp of perm * var


(*
type perm_sign = Pos | Neg
type perm = {sign : perm_sign; map : (atom*atom) AtomMap.t}
let id = {sign=Pos; map = AtomMap.empty}
let transpose a b = 
  let m = AtomMap.empty in
  let m = AtomMap.add a (b,b) m in
  let m = AtomMap.add b (a,a) m in
  {sign = Pos; map = m}
let compose p q = *)

type equations = (term*term) list

type subst = term Var.map
let empty = Var.empty
let lookup s x = 
  if Var.mem x s then Some(Var.find x s)
  else None
let join s s' = 
  Var.fold Var.add s s'
let mk_subst l = 
  List.fold_right (fun (x,t) s -> Var.add x t s) l empty

type constr = (atom * term) list

type prop = TermProp of sym * term
  | FreshProp of atom * term

type clause = prop * constr * prop list

type decl = AtomDecl of atom_sort
  | DataDecl of data_sort
  | PropDecl of sym * sort
  | SymDecl of sym * arity
  | ClauseDecl of clause 
  | Query of prop * prop list
  | UseDirective of string
  | TraceDirective of int



let rec apply_perm_atom p a = 
  match p with
    [] -> a
  | (b,c)::p' -> 
      let a' = apply_perm_atom p' a in
      if a' = b then c
      else if a' = c then b
      else a'

let rec apply_perm p t =
  let h = apply_perm p in
  match t with
    Atom a -> Atom(apply_perm_atom p a)
  | Unit -> Unit
  | Pair (e1,e2) -> Pair (h e1, h e2)
  | Abs (a,e) -> Abs(apply_perm_atom p a, h e)
  | App(f,e) -> App(f, h e)
  | Susp (p',x) -> Susp(compose p p', x)

let apply_perm_prop p q = 
  match q with
    TermProp (s,tm) -> TermProp(s,apply_perm p tm)
  | FreshProp (a,tm) -> FreshProp(apply_perm_atom p a, apply_perm p tm)

let rec apply_subst s t = 
  let h = apply_subst s in
  match t with
    Atom a -> Atom a
  | Abs (a,e) -> Abs(a, h e)
  | Pair(e1,e2) -> Pair(h e1, h e2)
  | App(f,e) -> App(f,h e)
  | Unit -> Unit
  | Susp(p,x) -> 
      (match lookup s x with 
	Some tm -> apply_perm p tm
      | None -> Susp(p,x))

let apply_subst_prop s p = 
  match p with
    TermProp (f,t) -> TermProp (f,apply_subst s t)
  | FreshProp (a,t) -> FreshProp(a,apply_subst s t)


let invert_perm p = 
  List.rev p

exception NYI
let rec domain_perm p = 
  match p with 
    [] -> []
  | (a,b)::p -> a::b::domain_perm p

let disagreement_set p1 p2 = 
  let dom1 = domain_perm p1 in
  let dom2 = domain_perm p2 in
  let dom = dom1 @ dom2 in
  let rng = List.map (fun x -> x, apply_perm_atom p1 x, apply_perm_atom p2 x) dom in
  List.map (fun (x,y,z) -> x) 
    (List.filter (fun (x,y,z) -> not(y = z)) rng)




let rec pp_perm oc p = 
  let out = output_string oc in
  match p with 
    [] -> ()
  | (a,b)::p -> out "("; out a; out " "; out b; out ")"

let rec pp_term oc term = 
  let out = output_string oc in
  let rec pp paren term = 
    match term with
      Unit -> out "()"
    | Atom a -> out "@"; out a
    | Pair (b,c) -> 
	if paren 
	then (out "("; pp true b; out ","; pp false c; out ")")
	else (pp true b; out ","; pp false c)
    | App (f,Unit) -> out f
    | App (f,(Pair(_,_) as c)) -> out f; pp true c
    | App (f,c) -> out f; out " "; pp false c
    | Abs (a,c) -> out "<"; out a; out ">"; pp true c
    | Susp (p,x) -> pp_perm oc p; out (Var.to_string x)
  in pp true term


let rec pp_sort oc s = 
  let out = output_string oc in
  let rec pp paren s = 
    match s with
      AtomSt(a) -> out "@"; out a
    | DataSt(a) -> out a
    | UnitSt -> out "unit"
    | PairSt(s1,s2) -> 
	if paren 
	then (out "("; pp true s1; out " * "; pp false s2; out ")")
	else (pp true s1; out " * "; pp false s2)
    | AbsSt(a,s) -> out "<"; out a; out ">"; pp true s
    | ListSt(s) -> pp true s; out " list"
  in pp false s

let rec pp_subst oc s = 
  let out = output_string oc in
  Var.iter 
    (fun v e -> out (Var.to_string v); out " -> "; pp_term oc e; out "\n";)
    s

let rec pp_constr oc c = 
  let out = output_string oc in
  match c with
    [] -> ()
  | (v,e)::c -> out v; out " # "; pp_term oc e; out "\n"; pp_constr oc c

let rec intersperse f g l = 
  match l with
    [] -> ()
  | [x] -> f x
  | x::xs -> f x; g(); intersperse f g xs

let pp_prop oc prop = 
  let out = output_string oc in
  match prop with
    TermProp (s,Unit) -> out s
  | TermProp (s,t) -> out s ; pp_term oc t
  | FreshProp (a,t) -> out a; out " # "; pp_term oc  t
;;

let pp_clause oc (p,cs,ps) = 
  let out = output_string oc in
  pp_prop oc p;
  
  if not(cs = []) 
  then 
    (out " / "; 
     intersperse (fun (a,tm) -> out a ; out " # "; pp_term oc tm) 
       (fun () -> out ",") cs);
  if not(ps = [])
  then 
    (out " <- "; 
     intersperse (pp_prop oc) (fun () -> out ", ") ps)
	

let pp_decl oc decl = 
  let out = output_string oc in
  match decl with
    AtomDecl an -> out an; out " : atom."
  | DataDecl dn -> out dn; out " : type."
  | PropDecl (sym,UnitSt) -> out sym; out " : "; out " prop."
  | PropDecl (sym,s) -> out sym; out " : "; pp_sort oc s; out " -> prop."
  | SymDecl (sym,(UnitSt,t)) -> out sym; out " : "; out t; out "."
  | SymDecl (sym, (s,t)) -> out sym; out " : "; pp_sort oc s; out " -> "; out t; out "."
  | ClauseDecl clause  -> pp_clause oc clause ; out "."
  | Query (p,ps) -> 
      pp_prop oc p; out " ?- "; intersperse (pp_prop oc) (fun () -> out ", ") ps; out "."
  | UseDirective(f) -> 
      out "#use \"" ; out f; out "\"."


let rec fvs term = 
  match term with
    Atom a -> []
  | Unit -> []
  | Pair(e1,e2) -> fvs e1 @ fvs e2
  | App(f,e) -> fvs e
  | Abs(a,e) -> fvs e
  | Susp(p,x) -> [x] (* ??? what if there are permutations? is this sound? *)

let fvs_prop p = 
  match p with
    TermProp (f,t) -> fvs t
  | FreshProp (a,t) -> fvs t



let rename p ps = 
  let rec ren s tm = 
    match tm with
    | Atom a -> s,Atom a
    | Unit -> s,Unit
    | Pair(e1,e2) -> 
	let s,t1 = ren s e1 in 
	let s,t2 = ren s e2 in
	s, Pair(t1,t2)
    | Abs(a,e) -> 
	let s,t = ren s e in
	s, Abs(a,t)
    | App(f,e) -> 
	let s,t = ren s e in
	s, App(f,t)
    | Susp(p,x) -> 
	if Var.mem x s 
	then s,Susp(p,Var.find x s)
	else 
	  let y = Var.rename x in
	  Var.add x y s, Susp(p,y)
  in
  let rec go s l = 
    match l with [] -> s,[]
    | TermProp(f,t)::tms -> 
	let s,t = ren s t in
	let s,tms = go s tms in
	s,(TermProp(f,t)::tms)
    | FreshProp(a,t)::tms -> 
	let s,t = ren s t in
	let s,tms = go s tms in
	s,(FreshProp(a,t)::tms)
  in 
  let s,[p] = go Var.empty [p] in
  let _,ps = go s ps in
  p,ps


let atom_occurs_in a t = 
  let rec h t = 
    match t with
      Atom b -> a = b
    | Susp(p,x) -> false (*??? don't care here... what's best?*)
    | Pair(t1,t2) -> h t1 && h t2
    | Abs(b,t) -> if a = b then true else h t
    | Unit -> false
    | App(f,t) -> h t
  in h t

let rec sort_to_string s = 
  let out s = s in
  let pp = sort_to_string in 
  match s with
    AtomSt(a) -> out "@" ^ out a
  | DataSt(a) -> out a
  | UnitSt -> out "unit"
  | PairSt(s1,s2) -> out "("^ pp s1^ out " * "^ pp s2^ out ")"
  | AbsSt(a,s) -> out "<"^ out a^ out ">"^ pp s
  | ListSt(s) -> out "("^ pp s^ out ") list"
	
