open Absyn;;

type ctx = sort Var.map 
type actx = (string * atom_sort) list;;
exception TcFail of string;;

let force_symk sg s =
  try match List.assoc s sg with 
    SymK ar -> ar
  | _ -> raise (TcFail ("expected "^s^" to be a function symbol"))
  with Not_found -> raise (TcFail ("unbound symbol "^s))

let force_atomsk sg s =
  try match List.assoc s sg with 
    AtomSK -> ()
  | _ -> raise (TcFail ("expected "^s^" to be an atom sort"))
  with Not_found -> raise (TcFail ("unbound symbol "^s))
	  
let force_datask sg s =
  try match List.assoc s sg with 
    DataSK -> ()
  | _ -> raise (TcFail ("expected "^s^" to be a data sort"))
  with Not_found -> raise (TcFail ("unbound symbol "^s))

let force_atomk sg s =
  try match List.assoc s sg with
    AtomK ats -> ats
  | _ -> raise (TcFail ("expected "^s^" to be an atom"))
  with Not_found -> raise (TcFail ("unbound symbol "^s))

let force_propk sg s = 
  try match List.assoc s sg with
    PropK st -> st 
  | _ -> raise (TcFail ("expected "^s^" to be a prop symbol"))
  with Not_found -> raise (TcFail ("unbound symbol "^s))

let rec check_sort_eq s1 s2 = 
  match s1, s2 with
    AtomSt ats, AtomSt ats' -> ats = ats'
  | DataSt dts, DataSt dts' -> dts = dts'
  | UnitSt, UnitSt -> true
  | PairSt(s1,s2), PairSt(s1',s2') -> 
      check_sort_eq s1 s1' && check_sort_eq s2 s2'
  | AbsSt(ats,s), AbsSt(ats',s') -> 
      ats = ats' && check_sort_eq s s'
  | ListSt s, ListSt s' -> 
      check_sort_eq s s'
  | _ -> false

let rec check_sort sg s = 
  match s with 
    UnitSt -> ()
  | AtomSt(ats) -> force_atomsk sg ats
  | DataSt(dts) -> force_datask sg dts
  | PairSt(s1,s2) -> check_sort sg s1; check_sort sg s2
  | AbsSt(ats,s) -> force_atomsk sg ats; check_sort sg s
  | ListSt s -> check_sort sg s

let check_atom_sort actx a ats = 
  try let ats' = List.assoc a actx in
  if ats <> ats' 
  then raise (TcFail ("atom "^a^" used with incompatible sorts "
	      ^ats^" and "^ats'^"."));
  actx
  with Not_found -> (a,ats)::actx

let check_var_sort ctx x s = 
  try let s' = Var.find x ctx in
  if not(check_sort_eq s s')
  then raise (TcFail ("variable "^Var.to_string x^" used with incompatible sorts "^sort_to_string s^" and "^sort_to_string s'^"."));
  ctx
  with Not_found -> Var.add x s ctx

let rec check_term sg ctx actx tm s = 
  match tm,s with
    Atom a, AtomSt(ats) -> 
      ctx, check_atom_sort actx a ats
  | Unit,UnitSt -> ctx,actx
  | Pair(t1,t2),PairSt(s1,s2) ->
      let ctx1,actx1 = check_term sg ctx actx t1 s1 in
      let ctx2,actx2 = check_term sg ctx1 actx1 t2 s2 in
      ctx2,actx2
  | App(sym,term),DataSt(ds) -> 
      let (s',ds') = force_symk sg sym in
      if ds <> ds' then raise (TcFail ("expected "^ds^", got "^ds'^"."));
      check_term sg ctx actx term s'
  | Abs(a,term), AbsSt(ats,s) ->
      let actx = check_atom_sort actx a ats in
      let ctx,actx = check_term sg ctx actx term s in
      ctx,actx
  | Susp(p,x),s ->  (* TODO: Check sorts of atoms in perm *)
      check_var_sort ctx x s, actx
  | Unit, ListSt s -> ctx,actx
  | Pair(t1,t2), ListSt s -> 
      let ctx,actx = check_term sg ctx actx t1 s in
      let ctx,actx = check_term sg ctx actx t2 (ListSt s) in
      ctx,actx
  | _ -> raise (TcFail ("term is incompatible with sort "^sort_to_string s))

let check_arity sg (s,dts) = 
  check_sort sg s;
  force_datask sg dts

let check_prop sg ctx actx prop = 
  match prop with
    TermProp(p,tm) -> 
      let s = force_propk sg p in
      check_term sg ctx actx tm s
  | FreshProp(a,tm) -> (* TODO: Check the term is WF somehow. *)
      ctx,actx

let check_constr sg ctx actx constr = 
  (* TODO: Check these really *)
  ctx,actx

let check_clause sg (p,cs,ps) = 
  let ctx,actx = check_prop sg Var.empty [] p in
  let ctx,actx = check_constr sg ctx actx cs in
  let ctx,actx = 
    List.fold_left (fun (ctx,actx) p -> check_prop sg ctx actx p) (ctx,actx) ps in
  ()
    

let check_decl sg decl = 
  match decl with 
    AtomDecl atomsort -> 
      (atomsort,AtomSK)::sg
  | DataDecl datasort ->       
      (datasort,DataSK)::sg
  | PropDecl (p,sort) -> 
      check_sort sg sort;
      (p,PropK sort)::sg
  | SymDecl (f,arity) -> 
      check_arity sg arity;
      (f,SymK arity)::sg
  | ClauseDecl c -> check_clause sg c; sg
  | Query (p,ps) -> 
      let ctx,actx = check_prop sg Var.empty [] p in
      let ctx,actx = List.fold_left (fun (ctx,actx) p -> check_prop sg ctx actx p) (ctx,actx) ps in
      sg
  | UseDirective _ -> sg      
  | TraceDirective _ -> sg      
