open Absyn;;
open Solve;;

exception Error of string
(* Is there a smart way to deal with this using atoms? *)
(* TODO: Implement unification relative to quantifier context,    which obviates this *)
let skolemize tm tms = 
  let fvs = List.concat (List.map Absyn.fvs_prop tms) in
  let skolem  = mk_subst (List.map (fun x -> (x,App("skolem#"^(Var.to_string x),Unit))) fvs) in
  let tm' = apply_subst_prop skolem tm in
  let tms' = List.map (apply_subst_prop skolem) tms in
  tm',  (List.map (fun tm -> tm,[],[]) tms');;


(* TODO: Implement qctxts and printing relative to substitution,    make this unnecessary *)

let unravel fvs subst = 
  let rec h e = 
    match e with
      Atom a -> Atom a
    | Unit -> Unit
    | Pair(e1,e2) -> Pair(h e1, h e2) 
    | Abs(a,e) -> Abs(a,h e)
    | App(f,e) -> App(f, h e)
    | Susp(p,x) -> 
	(match lookup subst x with 
	  Some e' -> 
	    let e'' = apply_perm p e' in
	    h e''
	| None -> Susp(p,Var.squash x))
  in mk_subst (List.map (fun x -> (x,h (Susp(id,x)))) fvs);;

let get_decls file = 
  if Sys.file_exists file then
    let decls = 
      let ic = open_in file in
      Lexer.lineno := 0;
      try Parser.parse Lexer.lex (Lexing.from_channel ic) 
      with Parsing.Parse_error -> 
	raise (Error("Parse error near line "^(string_of_int (!Lexer.lineno))^"\n"))  in
    decls
  else (print_string (file^": File not found\n"); [])
      
let interactive = ref false
      
let rec run sg clauses  rest = 
  match rest with
    [] -> sg,clauses
  | (Query (tm,tms) as decl)::rest -> 
      let sg = Tc.check_decl sg decl in
      let tm', tmclauses = skolemize tm tms in 
      let fvs = Absyn.fvs_prop tm' in
      print_string "--------\nQuery: ";
      pp_decl stdout decl;
      print_string "\n";
      let curclauses = (List.rev(clauses)@tmclauses) in
      let stk = [[tm'], curclauses, empty,[]] in
      Solve.reset_atom();
      Var.reset_var();
      let rec do_solve stk = 
	(try 
	let subst,constrs,stk = Solve.solve curclauses stk in
	let newsubst = unravel fvs subst in
	print_string "\nQuery solved:\n"; 
	print_string "{\n";
	pp_subst stdout newsubst;
	pp_constr stdout constrs;
	print_string "}\n";
	print_string "--------\n";
	flush stdout;
	if !interactive 
	then 
	  let s = input_line stdin in
	  if s = ";" then do_solve stk
	with Fail -> print_string "\nQuery failed\n---------\n"; 
	            flush stdout) in
      do_solve stk;
      run sg clauses rest
  | (ClauseDecl(c) as decl)::rest -> 
      (try Tc.check_clause sg c;
      with Tc.TcFail msg -> 
	raise (Error(msg ^ " in clause \nTODO: stringize clauses\n")));
      run sg (c::clauses) rest
  | (UseDirective(file))::rest -> 
      let old_interactive = !interactive in
      interactive := false;
      let decls = get_decls file in
      let sg',clauses' = run sg clauses decls in
      interactive := old_interactive;
      run sg' clauses' rest
  | (TraceDirective(i))::rest -> 
      Solve.trace := i;
      run sg clauses rest
  | decl::rest -> 
      let sg = Tc.check_decl sg decl in
      pp_decl stdout decl;
      print_string "\n";
      run sg clauses rest;;
	
       
let lexbuf = Lexing.from_channel stdin
let rec toploop sg decls = 
  try 
    print_string "nolog> ";
    flush stdout;
    Lexer.lineno := 0;
    try let result = Parser.parse_line Lexer.lex lexbuf in
    if result = [] then exit(0);
    let sg',decls' = run sg decls result in
    toploop sg' decls'
    with Parsing.Parse_error -> 
      print_string ("Parse error near line "
		    ^(string_of_int (!Lexer.lineno))^"\n");
      toploop sg decls
    | Error  msg -> 
	print_string msg;
	toploop sg decls
    | Tc.TcFail msg -> 
	print_string ("Type error: "^msg);
	toploop sg decls

  with 
  | _ -> ()
      
let main args = 
  let timer = ref 0.0 in
  let start_timer () = 
    timer := Sys.time() in
  let time_msg msg = 
    let diff = Sys.time() -. !timer in
    print_string (msg^" took "^(string_of_float diff)^" s\n");
    flush stdout
  in
  let rsg = ref [] in
  let rdecls = ref [] in
  let do_file file = 
    start_timer();
    let decls = get_decls file in
    time_msg "Parsing";
    start_timer();
    let sg,decls = run !rsg !rdecls decls in
    rsg := sg;
    rdecls := decls;
    time_msg "Running"
  in 
   let spec = [("-t", Arg.Int(fun i -> Solve.trace := i), "Enable tracing")] in
   print_string "Nolog 0.1\n";
   flush stdout;
   Arg.parse spec do_file "usage: nolog [-t] files";
   interactive := true;
   toploop !rsg !rdecls;;
   

main Sys.argv;;
